#!/usr/bin/perl
#
# $Id$
#
# (C) IBCL Informatik-Buero e.U. 2010
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

use strict;

use Dpkg::Version;

require "/usr/lib/pba/pba-functions.pl";


sub prepareGetopts() {

  # $main::helpMessage must be set by bin/pba-*
  if ($main::optionList eq "") {
    $main::optionList = "dc:o:qv";
  }
 
  $main::role = "devtools";
}


sub initDevToolsFunctions() {

  $main::configParams{"GPGKeyId"} = "";
  $main::configParams{"DefaultBuildHost"} = "";

  $main::directorySelector = $main::opts{"o"};

  $main::demandedBuildHost = $main::opts{"b"};

  $main::configFile = "/etc/pba/devtools.conf";
  $main::userConfigFile = $main::userHome."/.pba/devtools.conf";

}

sub checkDevToolsConfig() {

  $main::gpgHome = ""; # make sure this is defined empty
  $main::gpgKeyId = $main::configParams{"GPGKeyId"};
  if ($main::gpgKeyId eq "") {
    $main::dpkgGpgOptions = "-us -uc";
    ($main::verbosity > 0) && print "No GPG key configured, not signing dsc packages.\n";
  } else {
    if ($main::opts{"u"}) {
      $main::dpkgGpgOptions = "-us -uc";
      ($main::verbosity > 0) && print "Creating unsigned dsc packages despite GPG key, due to -u.\n";
    } else {
      $main::dpkgGpgOptions = "-k".$main::gpgKeyId;
      ($main::verbosity > 0) && print "Using GPG key with ID ".$main::gpgKeyId." to sign dsc packages.\n";
    }
  }

  $main::defaultBuildHost = $main::configParams{"DefaultBuildHost"};

  if ($main::demandedBuildHost eq "DEFAULT") {
    if ($main::defaultBuildHost ne "") {
      $main::demandedBuildHost = $main::defaultBuildHost;
      print "\nInfo: Called with -b DEFAULT, setting demandedBuildHost to DefaultBuildHost, ".$main::demandedBuildHost."\n\n";
    } else {
      print "\nWarning: Called with -b DEFAULT, but no DefaultBuildHost configured in devtools.conf, ignoring -b parameter.\n\n";
      $main::demandedBuildHost = "";
    }
  } elsif ($main::demandedBuildHost eq "AUTO") {
    if ($main::defaultBuildHost ne "") {
      $main::demandedBuildHost = $main::defaultBuildHost;
      print "\nInfo: Called with -b AUTO, setting demandedBuildHost to DefaultBuildHost, ".$main::demandedBuildHost."\n\n";
    } else {
      my $nrBuildHosts = scalar(@main::buildhosts);
      if ($nrBuildHosts == 0) {
        print "\nError: Called with -b AUTO, but no buildhosts configured in devtools.conf, ignoring -b parameter and going to die later.\n\n";
      } else {
        my $randomBuildHostNr = int(rand($nrBuildHosts));
        $main::demandedBuildHost = @main::buildhosts[$randomBuildHostNr];
        print "\nInfo: Called with -b AUTO, setting demandedBuildHost to randomly selected buildhost ".$main::demandedBuildHost."\n\n";
      }
    }
  }

}


sub readDebianFiles() {

  my $selectorPattern = 'debian.*'.$main::directorySelector;

  @main::debianDirs = ();
  @main::controlFiles = ();
  @main::changelogFiles = ();

  # find debian*/control
  my @dirs = glob("debian*");
  # my @dirs = glob("*");
  for my $dir (@dirs) {
    ($main::verbosity > 2) && print "Checking ".$dir."\n";
    if (! -d $dir) {
      ($main::verbosity > 1) && print "Skipping non-directory ".$dir."\n";
      next; 
    }
    if ($dir !~ /^debian/) {
      ($main::verbosity > 1) && print "Skipping non-debian file ".$dir."\n";
      next; 
    }
    if (! -f $dir."/control") {
      ($main::verbosity > 1) && print "ERROR: Skipping directory without control file: ".$dir."\n";
      next; 
    }
    if (! -f $dir."/changelog") {
      ($main::verbosity > 1) && print "ERROR: Skipping directory without changelog file: ".$dir."\n";
      next; 
    }
    if ($main::directorySelector ne "") {
      ($main::verbosity > 2) && print "Checking against selector '".$main::directorySelector."', pattern '".$selectorPattern."'\n";
      if ($dir !~ /$selectorPattern/) {
        #if ($main::scriptName =~ /^pba-rpm/) {
        #  ($main::verbosity > 2) && print "Directory ".$dir." does not match selector '".$main::directorySelector."', in pba-rpm* command reading it anyway for consistency.\n";
        #} else {
        #  ($main::verbosity > 2) && print "Skipping directory that does not match selector '".$main::directorySelector."': ".$dir."\n";
        #  next; 
        #}
        ($main::verbosity > 2) && print "Directory ".$dir." does not match selector '".$main::directorySelector."', reading it anyway for consistency checks.\n";
      }
    }
    ($main::verbosity > 2) && print "Using ".$dir."\n";
    push @main::debianDirs, $dir;
    push @main::controlFiles, $dir."/control";
    push @main::changelogFiles, $dir."/changelog";
  } 

  %main::buildDepsByControlFile = ();
  $main::buildDeps = "";

  %main::instDepsByControlFile = ();
  $main::instDeps = "";

  %main::sourcePackageNamesByControlFile = ();
  @main::sourcePackageNames = "";
  #%main::sourcePackageArchitecturesByPackage = ();

  %main::binaryPackageNamesByControlFile = ();
  %main::debPackageNamesByControlFile = ();
  %main::debDbgSymPackageNamesByControlFile = ();
  %main::repositoriesByControlFile = ();

  @main::binaryPackageNames = "";
  %main::binaryPackageArchitecturesByPackage = ();
  
  %main::pbaDistributionListByControlFile = ();

  my $currentSourcePackageName = "";
  my $currentBinaryPackageName = "";

  my $inBuildDeps = 0;
  my $inInstDeps = 0;
  my $currentBuildDeps = "";
  my $currentInstDeps = "";

  if (scalar(@main::controlFiles) == 0) {
    if ($main::scriptName =~ /^pba-rpm/) {
      print("WARNING: No debian*/control files to parse. Continuing anyway under pba-rpm* command.\n");
    } else {
      cleanupAndDie("FATAL: No control files to parse.\n");
    }
  }

  # parse debian*/control
  foreach my $controlFile (@main::controlFiles) {
    $currentSourcePackageName = "";
    $currentBinaryPackageName = "";
    open (CONTROLFILE, "<".$controlFile) || cleanupAndDie("FATAL: Cannot read ".$controlFile."\n");
    my ($debDir, $controlFileBasename) = split /\//, $controlFile, 2;
    while ( my $curLine = <CONTROLFILE> ) {
      chomp $curLine;
      if ($inBuildDeps) {
        if ($curLine =~ /^\s/) {
          $currentBuildDeps .= $curLine;
          next;
        } else { 
          #print "DEBUG: full currentBildDebs: '".$currentBuildDeps."'\n";
          $currentBuildDeps = simplifyList($currentBuildDeps);
          #print "DEBUG: currentBildDebs after simplyfiylist: '".$currentBuildDeps."'\n";
          # remove version restrictions
          $currentBuildDeps =~ s/ ?\([^\(\)]+\)//g;
          #print "DEBUG: currentBildDebs after version stripping: '".$currentBuildDeps."'\n";
          # remove alternatives
          $currentBuildDeps =~ s/[^ ]+\|[^ ]+ ?//g;
          #print "DEBUG: currentBildDebs after alternative stripping: '".$currentBuildDeps."'\n";
          $main::buildDepsByControlFile{$controlFile} = $currentBuildDeps;
          if (($main::directorySelector eq "") or ($debDir =~ /$selectorPattern/)) {
            $main::buildDeps .= $currentBuildDeps . " ";
          } else {
            ($main::verbosity > 2) && print "Not adding builddeps from ".$debDir." (not matching selector '".$main::directorySelector."') to global builddeps list\n";
          }
          $inBuildDeps = 0;
          $currentBuildDeps = "";
          # continue to analyse the line!
        }
      }
      if ($inInstDeps) {
        if ($curLine =~ /^\s/) {
          $currentInstDeps .= $curLine;
          next;
        } else { 
          $currentInstDeps = simplifyList($currentInstDeps);
          # remove version restrictions
          $currentInstDeps =~ s/ ?\([^\(\)]+\)//g;
          # remove alternatives
          $currentInstDeps =~ s/[^ ]+\|[^ ]+ ?/ /g;
          $main::instDepsByControlFile{$controlFile} = $currentInstDeps;
          if (($main::directorySelector eq "") or ($debDir =~ /$selectorPattern/)) {
            $main::instDeps .= $currentInstDeps . " ";
          } else {
            ($main::verbosity > 2) && print "Not adding instdeps from ".$debDir." (not matching selector '".$main::directorySelector."') to global instdeps list\n";
          }
          $inInstDeps = 0;
          $currentInstDeps = "";
          # continue to analyse the line!
        }
      }
      my ( $pn, $pv ) = split /: */, $curLine, 2;
      if ($pn =~ /^Build-Depends$/) {
        $inBuildDeps = 1;
        $currentBuildDeps = $pv;
        next;
      }
      if ($pn =~ /^Depends$/) {
        $inInstDeps = 1;
        $currentInstDeps .= " ".$pv;
        next;
      }
      if ($pn =~ /^Source$/) {
        $currentSourcePackageName = $pv;
        $main::sourcePackageNamesByControlFile{$controlFile} = $pv;
        if (($main::directorySelector eq "") or ($debDir =~ /$selectorPattern/)) {
          push @main::sourcePackageNames, $pv;
        }
        next;
      }
      if ($pn =~ /^Package$/) {
        $currentSourcePackageName = "";
        $currentBinaryPackageName = $pv;
        if (defined $main::binaryPackageNamesByControlFile{$controlFile}) {
          $main::binaryPackageNamesByControlFile{$controlFile} .= " ".$pv;
        } else {
          $main::binaryPackageNamesByControlFile{$controlFile} = $pv;
        }
        if (($main::directorySelector eq "") or ($debDir =~ /$selectorPattern/)) {
          push @main::binaryPackageNames, $pv;
        }
        next;
      }
      if ($pn =~ /^Architecture$/) {
        $main::binaryPackageArchitecturesByPackage{$currentBinaryPackageName} = $pv;
        #print "binPackage $currentBinaryPackageName has architecture ".$pv."\n";
        next;
      }
      if ($pn =~ /^X.*-PBA-Repositories$/i) {
        $main::repositoriesByControlFile{$controlFile} = simplifyList($pv);
        next;
      }
      if ($pn =~ /^X.*-PBA-Distributions$/i) {
        $main::pbaDistributionListByControlFile{$controlFile} = simplifyList($pv);
        next;
      }
    }  
    close CONTROLFILE;
  }

  %main::appVersionByChangelog = ();
  %main::upstreamVersionByChangelog = ();
  %main::releaseVersionByChangelog = ();
  %main::versionRestByChangelog = ();
  %main::packageNameByChangelog = ();
  %main::firstChangelogByChangelog = ();
  %main::firstLineByChangelog = ();

  # parse debian*/changelog
  foreach my $changelogFile (@main::changelogFiles) {
    $currentSourcePackageName = "";
    $currentBinaryPackageName = "";
    my $inFirstChangelog = 0;
    my $firstChangelog = "";
    my $versionRest = "";
    open (CONTROLFILE, "<".$changelogFile) || cleanupAndDie("FATAL: Cannot read ".$changelogFile."\n");
    while ( my $curLine = <CONTROLFILE> ) {
      chomp $curLine;
      if ($main::appVersionByChangelog{$changelogFile} eq "") {
        my $appVersion = $curLine;
        $appVersion =~ s/^.*\(//;
        $appVersion =~ s/\).*$//;
        $appVersion =~ s/^[0-9,\.]+://; # strip upstream: prefix
        if ($appVersion eq "") {
          cleanupAndDie("FATAL: No package version in ".$changelogFile."\n");
        } 
        my $upstreamVersion = "";
        my $releaseVersion = "";
        if ($appVersion =~ /-/) {
          ($upstreamVersion, $releaseVersion) = split /-/, $appVersion, 2;
        } else {
          my @appVersionParts = split /\./, $appVersion;
          $releaseVersion = pop @appVersionParts;
          $upstreamVersion = join(".", @appVersionParts);
        }
        if ($main::verbosity > 1) {
          print $changelogFile." has appVersion ".$appVersion.", upstreamVersion ".$upstreamVersion.", releaseVersion ".$releaseVersion."\n";
        }
        $main::appVersionByChangelog{$changelogFile} = $appVersion;
        $main::upstreamVersionByChangelog{$changelogFile} = $upstreamVersion;
        $main::releaseVersionByChangelog{$changelogFile} = $releaseVersion;
        my $packageName = $curLine;
        $packageName =~ s/ .*$//;
        if ($packageName eq "") {
          cleanupAndDie("FATAL: No package name in ".$changelogFile."\n");
        }
        $main::packageNameByChangelog{$changelogFile} = $packageName;
        my $versionRest = $curLine;
        $versionRest =~ s/^.*\) //;
        $main::versionRestByChangelog{$changelogFile} = $versionRest;
        $main::firstLineByChangelog{$changelogFile} = $curLine;
        $inFirstChangelog = 1;
        next;
      }
      if ($inFirstChangelog == 0) { 
        next; 
      }
      # collect first changelog entry
      if ($curLine =~ /^( |$)/) {
        $firstChangelog .= $curLine."\n";
        next;
      } 
      $inFirstChangelog = 0;
      #if ($main::verbosity > 2) {
      #  print $changelogFile." has this firstChangelog:\n".$firstChangelog."\n(end of firstChangelog)\n";
      #}
      $main::firstChangelogByChangelog{$changelogFile} = $firstChangelog;
    }  
    close CONTROLFILE;
    # FUTURE: list versions
    # FUTURE: compare versions
  }


  # check if source package names equals changelog package names
  # list expected packages by debianDir
  foreach  my $debianDir (@main::debianDirs) {
    if ($main::sourcePackageNamesByControlFile{$debianDir."/control"} ne $main::packageNameByChangelog{$debianDir."/changelog"}) {
      cleanupAndDie("FATAL: Source package name in ".$debianDir."/control, ".$main::sourcePackageNamesByControlFile{$debianDir."/control"}.",\ndiffers from source package name in ".$debianDir."/changelog, ".$main::packageNameByChangelog{$debianDir."/changelog"}.".");
    }
    if ($debianDir ne "debian") {
      if (-e "debian") {
        cleanupAndDie("FATAL: It's too dangerous to operate on ".$debianDir." while test -e debian is true.\n");
      }
    }
    my @binPackages = split / /, $main::binaryPackageNamesByControlFile{$debianDir."/control"};
    my $debPackages = "";
    my $debDbgSymPackages = "";
    foreach my $binPackageName (@binPackages) {
      if ($debPackages ne "") { $debPackages .= " "; }
      if ($debDbgSymPackages ne "") { $debDbgSymPackages .= " "; }
      #print "binPackage $binPackageName has architecture ".$main::binaryPackageArchitecturesByPackage{$binPackageName}."\n";
      if ($main::binaryPackageArchitecturesByPackage{$binPackageName} eq "all") {
         $debPackages .= "../".$binPackageName."_".$main::appVersionByChangelog{$debianDir."/changelog"}."_all.deb";
      } else {
         $debPackages .= "../".$binPackageName."_".$main::appVersionByChangelog{$debianDir."/changelog"}."_".myArchitecture().".deb";
         $debDbgSymPackages .= "../".$binPackageName."-dbgsym_".$main::appVersionByChangelog{$debianDir."/changelog"}."_".myArchitecture().".deb";
      }
    }
    $main::debPackageNamesByControlFile{$debianDir."/control"} = $debPackages;
    $main::debDbgSymPackageNamesByControlFile{$debianDir."/control"} = $debDbgSymPackages;
    if ($main::verbosity > 1) {
      print "Debian dir ".$debianDir." specifies source package ".$main::sourcePackageNamesByControlFile{$debianDir."/control"}."_".$main::appVersionByChangelog{$debianDir."/changelog"}.".dsc"."\n";
      print "which would produce these binary packages:\n";
      foreach my $debPackageName ($debPackages) {
        print "  ".$debPackageName."\n";
      }
      print "It also might produce these dbgsym packages:\n";
      foreach my $debPackageName ($debDbgSymPackages) {
        print "  ".$debPackageName."\n";
      }
    }
  }

}


sub buildBinaryPackages( $ ) {
  # $1 = debianDir
  my $debianDir = shift;
  ($debianDir eq "") && cleanupAndDie("FATAL: buildBinaryPackages called without debianDir as \$1.\n");

    ($main::verbosity > 0) && print "Going to build binary packages from ".$debianDir."/\n";

    # clean.sh?
    if (-x "clean.sh") { 
      ($main::verbosity > 1) && print "Calling ./clean.sh\n";
      executeOrDie("./clean.sh"); 
    } else {
      ($main::verbosity > 1) && print "No ./clean.sh to call.\n";
    }

    if ($debianDir ne "debian") {
      ($main::verbosity > 1) && print "Creating working copy debian/\n";
      if (-e "debian") {
        cleanupAndDie("FATAL: There already is a file or directory called debian, something must have gone wrong.\n");
      }
      executeOrDie("/bin/cp -a ".$debianDir." debian");
      if (! -d "debian") {
        cleanupAndDie("FATAL: /bin/cp -a ".$debianDir." debian did not create a debian/ directory\n");
      }
    }
    # we now have a debian/ directory to work on
    collectDevHostTxt();

    # create pseudo .dsc file if none is there, to satisfy stretch's buggy dpkg-buildpackage -b
    #my $dscFile = "../".$main::debPackageNamesByControlFile{$debianDir."/control"};
    my $dscFile = "../".$main::sourcePackageNamesByControlFile{$debianDir."/control"}."_".$main::appVersionByChangelog{$debianDir."/changelog"}.".dsc";
    if (-f $dscFile) {
      if ($main::verbosity > 1) { print "No need to touch dummy dsc file ".$dscFile."\n"; }
    } else {
      if ($main::verbosity > 1) { print "Touching dummy dsc file ".$dscFile."\n"; }
      printToFile($dscFile, "# dummy .dsc file");
      push @main::filesToCleanup, $dscFile;
    }

    my $dpkgCommand = "dpkg-buildpackage -b ".$main::dpkgGpgOptions;
    ($main::verbosity > 1) && print "Calling ".$dpkgCommand."\n";
    if (!execute($dpkgCommand)) { 
      my $msg;
      if ($debianDir ne "debian") {
        $msg = "Failed to build ".$debianDir.", please investigate and then /bin/rm -rf debian/\n";
      } else {
        $msg = "Failed to build ".$debianDir.", please investigate debian/\n";
      }
      cleanupAndDie($msg);
    }
    # list and remember created packages
    my @generatedPackages = split / /, $main::debPackageNamesByControlFile{$debianDir."/control"};
    my @generatedDbgSymPackages = split / /, $main::debDbgSymPackageNamesByControlFile{$debianDir."/control"};
    foreach my $generatedDbgSymPackage (@generatedDbgSymPackages) {
      if ( -f $generatedDbgSymPackage) {
        if ($main::verbosity > 2) {
          print "Found file $generatedDbgSymPackage, adding that to list of generatedPackages.\n"; 
        }
        push @generatedPackages, $generatedDbgSymPackage;
      } else {
        if ($main::verbosity > 2) {
          print "No file $generatedDbgSymPackage, not adding that to list of generatedPackages.\n"; 
        }
      }
    }
    #split / /, $main::debPackageNamesByControlFile{$debianDir."/control"};
    if ($main::verbosity > 0) {
      print "Done building binary packages from ".$debianDir."/.\n";
      print "To install these, call this:\nsudo dpkg -i";
      foreach my $generatedPackage (@generatedPackages) {
        print " \\\n  ".$generatedPackage;
      }
      print "\n\n";
    }  
    if ($debianDir ne "debian") {
      print "Cleaning up working copy debian/";
      executeOrDie("/bin/rm -rf debian");
    }
    print "Done building binary packages from ".$debianDir."/\n";
}


sub buildSourcePackage( $ ) {
  # $1 = debianDir
  my $debianDir = shift;
  ($debianDir eq "") && cleanupAndDie("FATAL: buildSourcePackage called without debianDir as \$1.\n");

    ($main::verbosity > 0) && print "Going to build sourcepackage from ".$debianDir."/\n";

    # clean.sh?
    if (-x "clean.sh") { 
      ($main::verbosity > 1) && print "Calling ./clean.sh\n";
      executeOrDie("./clean.sh"); 
    } else {
      ($main::verbosity > 1) && print "No ./clean.sh to call.\n";
    }

    if ($debianDir ne "debian") {
      ($main::verbosity > 1) && print "Creating working copy debian/\n";
      if (-e "debian") {
        cleanupAndDie("FATAL: There already is a file or directory called debian, something must have gone wrong.\n");
      }
      executeOrDie("/bin/cp -a ".$debianDir." debian");
      if (! -d "debian") {
        cleanupAndDie("FATAL: /bin/cp -a ".$debianDir." debian did not create a debian/ directory\n");
      }
    }
    # we now have a debian/ directory to work on
    collectDevHostTxt();
    my $dpkgCommand = "dpkg-buildpackage -S -d -I.svn -I.cproject -I.project -I.settings ".$main::dpkgGpgOptions;
    ($main::verbosity > 1) && print "Calling ".$dpkgCommand."\n";
    if (!execute($dpkgCommand)) { 
      my $msg;
      if ($debianDir ne "debian") {
        $msg = "Failed to build ".$debianDir.", please investigate and then /bin/rm -rf debian/\n";
      } else {
        $msg = "Failed to build ".$debianDir.", please investigate debian/\n";
      }
      cleanupAndDie($msg);
    }
    # list and remember created packages
    my $generatedPackage = "../".$main::sourcePackageNamesByControlFile{$debianDir."/control"}."_".$main::appVersionByChangelog{$debianDir."/changelog"}.".dsc";
    print "Done building source package ".$generatedPackage."\n";
    print "\n\n";
    if ($debianDir ne "debian") {
      print "Cleaning up working copy debian/";
      executeOrDie("/bin/rm -rf debian");
    } else {
      executeOrDie("/bin/rm -rf debian/devhost.txt.gz");
    }
    print "Done building source package from ".$debianDir."/\n";
}


sub buildPackages() {
  if (scalar(@main::controlFiles) > 0) {
    ($main::verbosity > 1) && print "Going to build all specifed debian dirs.\n";
  }
  
  my @allGeneratedPackages = ();
  my $selectorPattern = 'debian.*'.$main::directorySelector;
  
  foreach  my $debianDir (@main::debianDirs) {
    if (($main::directorySelector ne "") and ($debianDir !~ /$selectorPattern/)) {
      ($main::verbosity > 2) && print "Skipping directory that does not match selector '".$main::directorySelector."': ".$debianDir."\n";
      next;
    }
    buildBinaryPackages($debianDir);
    my @generatedPackages = split / /, $main::debPackageNamesByControlFile{$debianDir."/control"};
    my @generatedDbgSymPackages = split / /, $main::debDbgSymPackageNamesByControlFile{$debianDir."/control"};
    foreach my $generatedDbgSymPackage (@generatedDbgSymPackages) {
      if ( -f $generatedDbgSymPackage) {
        if ($main::verbosity > 2) {
          print "Found file $generatedDbgSymPackage, adding that to list of all generatedPackages, too.\n"; 
        }
        push @generatedPackages, $generatedDbgSymPackage;
      } else {
        if ($main::verbosity > 2) {
          print "No file $generatedDbgSymPackage, not adding that to list of all generatedPackages.\n"; 
        }
      }
    }
    push @allGeneratedPackages, @generatedPackages;
  }
  
  if (scalar(@main::controlFiles) > 1) {
    if ($main::verbosity > 0) {
      print "Done building all specified debian dirs.\n";
      print "To install all generated packages excute this:\nsudo dpkg -i";
      foreach my $generatedPackage (@allGeneratedPackages) {
        print " \\\n  ".$generatedPackage;
      }
    }
    print "\n\n";
  }
}


sub installPackages() {
  if (scalar(@main::controlFiles) > 1) {
    ($main::verbosity > 0) && print "Going to install packages from all specifed debian dirs.\n";
  }
  
  my @allGeneratedPackages = ();
  my $selectorPattern = 'debian.*'.$main::directorySelector;
  
  foreach  my $debianDir (@main::debianDirs) {
    if (($main::directorySelector ne "") and ($debianDir !~ /$selectorPattern/)) {
      ($main::verbosity > 2) && print "Skipping directory that does not match selector '".$main::directorySelector."': ".$debianDir."\n";
      next;
    }
    my @generatedPackages = split / /, $main::debPackageNamesByControlFile{$debianDir."/control"};
    my @generatedDbgSymPackages = split / /, $main::debDbgSymPackageNamesByControlFile{$debianDir."/control"};
    foreach my $generatedDbgSymPackage (@generatedDbgSymPackages) {
      if ( -f $generatedDbgSymPackage) {
        if ($main::verbosity > 2) {
          print "Found file $generatedDbgSymPackage, adding that to list of generatedPackages to install, too.\n"; 
        }
        push @generatedPackages, $generatedDbgSymPackage;
      } else {
        if ($main::verbosity > 2) {
          print "No file $generatedDbgSymPackage, not adding that to list of generatedPackages to install.\n"; 
        }
      }
    }
    push @allGeneratedPackages, @generatedPackages;
  }
 
  my $installCommand = "sudo dpkg -i"; 
  foreach my $generatedPackage (@allGeneratedPackages) {
    $installCommand .= " \\\n  ".$generatedPackage;
  }
  if ($main::verbosity > 0) {
    print "Going to call this:\n".$installCommand."\n";
  }
  executeOrDie($installCommand);
  if ($main::verbosity > 0) {
    print "Successfully installed these packages:\n  ".join(" \\\n  ",@allGeneratedPackages)."\n";
  }

}


sub releaseDebianPackages() {
  my $feedScript = "";

  if (scalar(@main::controlFiles) > 1) {
    ($main::verbosity > 0) && print "Going to release packages from all specifed debian dirs.\n";
  }

  my $releasedFilesDesc = "";

  my %buildHostByDebianDir = ();

  my $selectorPattern = 'debian.*'.$main::directorySelector;

  # first check prerequisites, evtl. pre-ask for build hosts
  foreach  my $debianDir (@main::debianDirs) {
    if (($main::directorySelector ne "") and ($debianDir !~ /$selectorPattern/)) {
      ($main::verbosity > 2) && print "Skipping directory that does not match selector '".$main::directorySelector."': ".$debianDir."\n";
      next;
    }
    my $controlFile = $debianDir."/control";
    if ($main::repositoriesByControlFile{$controlFile} eq "") {
      cleanupAndDie("FATAL: No XBCS-PBA-Repositories specified in ".$controlFile);
    }
    if ($main::demandedBuildHost ne "") {
      $buildHostByDebianDir{$debianDir} = $main::demandedBuildHost;
      print "\nSource package ".$main::sourcePackageNamesByControlFile{$controlFile}." from ".$debianDir."/ needs a build host.\n";
      if ((-d $main::demandedBuildHost) or ($main::demandedBuildHost =~ /\//)) {
        $main::demandedBuildHost =~ s/\/$//;
        print "Using the local directory specified with -b parameter: ".$buildHostByDebianDir{$debianDir}."\n";
        if (! -d $main::demandedBuildHost) {
          checkDirectory($main::demandedBuildHost);
        }
      } elsif (length($main::buildhostLongNames{$main::demandedBuildHost}) <= 0) {
        cleanupAndDie("FATAL: Buildhost specified with -b unknown: ".$main::demandedBuildHost);
      } else {
        print "Using the buildhost specified or hinted at with -b parameter: ".$buildHostByDebianDir{$debianDir}."\n";
      }
    } else  {
      if (scalar(@main::buildhosts) == 1) {
        $buildHostByDebianDir{$debianDir} = @main::buildhosts[0];
        print "\nSource package ".$main::sourcePackageNamesByControlFile{$controlFile}." from ".$debianDir."/ needs a build host.\n";
        print "Using the only configured buildhost: ".$buildHostByDebianDir{$debianDir}."\n";
      } else {
        if (scalar(@main::buildhosts) == 0) {
          cleanupAndDie("FATAL: No build hosts configured.");
        }
        # let user select buildhost
        print "\nSource package ".$main::sourcePackageNamesByControlFile{$controlFile}." from ".$debianDir."/ needs a build host. Configured build hosts are:\n\n";
        my $nrBuildHosts = scalar(@main::buildhosts);
        my $realDefaultBuildHost = $main::defaultBuildHost;
        my $defaultSelection = 1;
        if ($realDefaultBuildHost eq "") {
          $realDefaultBuildHost = @main::buildhosts[0];
        }
        for (my $i=1; $i<=$nrBuildHosts; $i++) {
          my $buildhost = @main::buildhosts[$i-1];
          if ($buildhost eq $realDefaultBuildHost) {
            print " ".$i."*: ".$buildhost."\n      ".$main::buildhostLongNames{$buildhost}."\n\n";
            $defaultSelection = $i;
          } else {
            print " ".$i.": ".$buildhost."\n      ".$main::buildhostLongNames{$buildhost}."\n\n";
          }
        }
        my $selection = 0;
        while ($selection < 1) {
          print "Which build shall I feed with packages from ".$debianDir."/? [".$defaultSelection."] ";
          my $userInput = <STDIN>;
          chomp $userInput;
          if ($userInput eq "") {
            $selection = $defaultSelection;
          } else {
            $selection = 0+$userInput;
          }
          if ( ($selection < 1) or ($selection > $nrBuildHosts) ) {
            print "Illegal input: ".$userInput."\nPlease try again!\n";
            $selection = 0;
          }
        }
        $buildHostByDebianDir{$debianDir} = @main::buildhosts[$selection-1];
        print "\nBuildhost selected for ".$debianDir." is: ".$buildHostByDebianDir{$debianDir}."\n";
      }
    }
  }

  foreach  my $debianDir (@main::debianDirs) {
    if (($main::directorySelector ne "") and ($debianDir !~ /$selectorPattern/)) {
      ($main::verbosity > 2) && print "Skipping directory that does not match selector '".$main::directorySelector."': ".$debianDir."\n";
      next;
    }
    my $controlFile = $debianDir."/control";
    ($main::verbosity > 0) && print "Going to release packages from ".$debianDir."/\n";

    buildSourcePackage($debianDir);
    my $generatedBasename = "../".$main::sourcePackageNamesByControlFile{$controlFile}."_".$main::appVersionByChangelog{$debianDir."/changelog"};
    my @generatedPackages = ();
    push @generatedPackages, $generatedBasename.".dsc";
    push @generatedPackages, $generatedBasename.".tar.gz";
    push @generatedPackages, $generatedBasename."_source.changes";
    
    #my $buildhost = @main::buildhosts[0];
    my $buildhost = $buildHostByDebianDir{$debianDir};
    if (-d $buildhost) {
      $feedScript .= proposeCpScript( $buildhost, join(" ",@generatedPackages) );
      $releasedFilesDesc .= "From ".$debianDir."/ to directory ".$buildhost."/:\n  ".join(" \\\n  ",@generatedPackages)."\n\n";
    } else {
      my $uploadName = $main::sourcePackageNamesByControlFile{$controlFile}."_".$main::appVersionByChangelog{$debianDir."/changelog"};
      my $feedHostDir = $main::buildhostFeedHostDirs{$buildhost};
      # FUTURE: add additional stuff referenced in .changes file
      $feedScript .= proposeFeedScript( $feedHostDir, $uploadName, join(" ",@generatedPackages) );
      $releasedFilesDesc .= "From ".$debianDir."/ to buildhost ".$buildhost.":\n  ".join(" \\\n  ",@generatedPackages)."\n\n";
    }
  }

  my $feedScriptFile = "/tmp/$$.sh";
  (-e $feedScriptFile) && unlink $feedScriptFile;
  $feedScript .= "\nexit 0\n";
  printToFile( $feedScriptFile, $feedScript, 0755 );
  if ($main::verbosity > 0) {
    print "Going to call feed script ".$feedScriptFile."\n";
  }
  executeOrDie($feedScriptFile);
  unlink $feedScriptFile;
  if ($main::verbosity > 0) {
    print "\nSuccessfully distributed these packages:\n\n".$releasedFilesDesc."\n";
  }

}


sub readSpecFiles() {

  my $selectorPattern = 'spec.*'.$main::directorySelector;

  @main::specDirs = ();
  @main::specFiles = ();

  # find spec/*.spec

  my @dirs = glob("spec*");
  # my @dirs = glob("*");
  for my $dir (@dirs) {
    ($main::verbosity > 2) && print "Checking ".$dir."\n";
    if (! -d $dir) {
      ($main::verbosity > 1) && print "Skipping non-directory ".$dir."\n";
      next; 
    }
    if ($dir !~ /^spec/) {
      ($main::verbosity > 1) && print "Skipping non-spec file ".$dir."\n";
      next; 
    }
    if ($main::directorySelector ne "") {
      ($main::verbosity > 2) && print "Checking against selector '".$main::directorySelector."', pattern '".$selectorPattern."'\n";
      if ($dir !~ /$selectorPattern/) {
        #($main::verbosity > 2) && print "Skipping directory that does not match selector '".$main::directorySelector."': ".$dir."\n";
        #next; 
        ($main::verbosity > 2) && print "Directory ".$dir." does not match selector '".$main::directorySelector."', reading it anyway for consistency checks.\n";
      }
    }
    ($main::verbosity > 2) && print "Candidating ".$dir."\n";
    my $dirSpecFiles = "";
    for my $controlFile (keys (%main::sourcePackageNamesByControlFile)) {
      my $specFileName = $dir."/".$main::sourcePackageNamesByControlFile{$controlFile}.".spec";
      if (-f $specFileName) {
        ($main::verbosity > 2) && print "Found ".$specFileName."\n";
        $dirSpecFiles .= $specFileName;
        push @main::specFiles, $specFileName;
      } else {
        ($main::verbosity > 2) && print "No ".$specFileName."\n";
      }
    } 
    if ($dirSpecFiles ne "")
    { push @main::specDirs, $dir; }
  }

  %main::versionsBySpecFile = ();
  %main::releasesBySpecFile = ();
  %main::buildHostsBySpecFile = ();
  %main::appVersionsBySpecFile = ();

  # check stuff Name in specFile?
  foreach my $specFile (@main::specFiles) {
    my ($specDir, $specFileName) = split /\//, $specFile, 2;
    my ($specBase, $specRest) = split /\./, $specFileName, 2;
    my $specName = "";
    $main::versionsBySpecFile{$specFile} = "";
    $main::releasesBySpecFile{$specFile} = "";
    open (SPECFILE, "<".$specFile) || cleanupAndDie("FATAL: Cannot read ".$specFile."\n");
    ($main::verbosity > 2) && print "Parsing ".$specFileName.", dir=".$specDir.", basename=".$specBase."\n";
    my $doneParsing = 0;
    while ( my $curLine = <SPECFILE> ) {
      if ($doneParsing) {
        next;
      }
      chomp $curLine;
      if ($curLine =~ /^\s*$/) {
        ($main::verbosity > 2) && print "Done parsing ".$specFileName."\n";
        $doneParsing = 1;
        next;
      }
      my ( $pn, $pv ) = split /: */, $curLine, 2;
      if ($pn =~ /^Name$/) {
        $specName = $pv;
        next;
      }
      if ($pn =~ /^Version$/) {
        $main::versionsBySpecFile{$specFile} = $pv;
        next;
      }
      if ($pn =~ /^Release$/) {
        $main::releasesBySpecFile{$specFile} = $pv;
        next;
      }
      if ($pn =~ /^# PBA-BuildHost$/) {
        $main::buildHostsBySpecFile{$specFile} = $pv;
        next;
      }
    }
    close (SPECFILE);
    if ($specName eq "")
    { cleanupAndDie("FATAL: No Name field in ".$specFile."\n"); }
    if ($specName ne $specBase)
    { print "WARNING: Name field in ".$specFile." should be ".$specBase.", is ".$specName."\n"; }
    if ($main::versionsBySpecFile{$specFile} eq "")
    { cleanupAndDie("FATAL: No Version field in ".$specFile."\n"); }
    if ($main::releasesBySpecFile{$specFile} eq "")
    { cleanupAndDie("FATAL: No Release field in ".$specFile."\n"); }
    $main::appVersionsBySpecFile{$specFile} = $main::versionsBySpecFile{$specFile} . "-" . $main::releasesBySpecFile{$specFile};
  }

}


sub releaseSRPMPackages() {

  if (scalar(@main::specFiles) > 1) {
    ($main::verbosity > 0) && print "Going to release packages from all specifed spec/*.spec files.\n";
  }

  # clean.sh?
  if (-x "clean.sh") { 
    ($main::verbosity > 1) && print "Calling ./clean.sh\n";
    executeOrDie("./clean.sh"); 
  } else {
    ($main::verbosity > 1) && print "No ./clean.sh to call.\n";
  }

  my $feedScript = "";
  my $releasedFilesDesc = "";

  my $tmpDir = "/tmp/".$$;
  if ( -d $tmpDir ) { executeOrDie("/bin/rm -rf ".$tmpDir); }
 
  executeOrDie("/bin/mkdir -pv ".$tmpDir);
  #push @main::dirsToCleanup, $tmpDir; 

  my $selectorPattern = 'spec.*'.$main::directorySelector;

  foreach my $specFile (@main::specFiles) {
    my ($specDir, $specFileName) = split /\//, $specFile, 2;
    if (($main::directorySelector ne "") and ($specDir !~ /$selectorPattern/)) {
      ($main::verbosity > 2) && print "Skipping spec directory that does not match selector '".$main::directorySelector."': ".$specDir."\n";
      next;
    } 
    my ($specBase, $specRest) = split /\./, $specFileName, 2;
    # future: check repository?
    if ($main::demandedBuildHost ne "") {
      $main::buildHostsBySpecFile{$specFile} = $main::demandedBuildHost;
      print "\nSource package ".$specBase." from ".$specFile." needs a build host.\n";
      if ((-d $main::demandedBuildHost) or ($main::demandedBuildHost =~ /\//)) {
        $main::demandedBuildHost =~ s/\/$//;
        print "Using the local directory specified with -b parameter: ".$main::demandedBuildHost."\n";
        if (! -d $main::demandedBuildHost) {
          checkDirectory($main::demandedBuildHost);
        }
      } elsif (length($main::buildhostLongNames{$main::demandedBuildHost}) <= 0) {
        cleanupAndDie("FATAL: Buildhost specified with -b unknown: ".$main::demandedBuildHost);
      } else {
        print "Using the buildhost specified or hinted at with -b parameter: ".$main::buildHostsBySpecFile{$specFile}."\n";
      }
    } else {
      if (scalar(@main::buildhosts) == 1) {
        $main::buildHostsBySpecFile{$specFile} = @main::buildhosts[0];
        print "\nSource package ".$specBase." from ".$specFile." needs a build host.\n";
        print "Using the only configured buildhost: ".$main::buildHostsBySpecFile{$specFile}."\n";
      } else {
        if (scalar(@main::buildhosts) == 0) {
          cleanupAndDie("FATAL: No build hosts configured.");
        }
        if ($main::buildHostsBySpecFile{$specFile} ne "") {
          print "\nSource package ".$specBase." from ".$specFile." needs a build host.\n";
          print "Using the buildhost specified in the .spec file: ".$main::buildHostsBySpecFile{$specFile}."\n";
        } else {
          # let user select buildhost
          print "\nSource package ".$specBase." from ".$specFile." needs a build host. Configured build hosts are:\n\n";
          my $nrBuildHosts = scalar(@main::buildhosts);
          my $realDefaultBuildHost = $main::defaultBuildHost;
          my $defaultSelection = 1;
          if ($realDefaultBuildHost eq "") {
            $realDefaultBuildHost = @main::buildhosts[0];
          }
          for (my $i=1; $i<=$nrBuildHosts; $i++) {
            my $buildhost = @main::buildhosts[$i-1];
            if ($buildhost eq $realDefaultBuildHost) {
              print " ".$i."*: ".$buildhost."\n      ".$main::buildhostLongNames{$buildhost}."\n\n";
              $defaultSelection = $i;
            } else {
              print " ".$i.": ".$buildhost."\n      ".$main::buildhostLongNames{$buildhost}."\n\n";
            }
          }
          my $selection = 0;
          while ($selection < 1) {
            print "Which build shall I feed with packages from ".$specFile."? [".$defaultSelection."] ";
            my $userInput = <STDIN>;
            chomp $userInput;
            if ($userInput eq "") {
              $selection = $defaultSelection;
            } else {
              $selection = 0+$userInput;
            }
            if ( ($selection < 1) or ($selection > $nrBuildHosts) ) {
              print "Illegal input: ".$userInput."\nPlease try again!\n";
              $selection = 0;
            }
          }
          $main::buildHostsBySpecFile{$specFile} = @main::buildhosts[$selection-1];
          print "\nBuildhost selected for ".$specFile." is: ".$main::buildHostsBySpecFile{$specFile}."\n";
        }
      }
    } # end if have buildhosts
    my $buildhost = $main::buildHostsBySpecFile{$specFile};
    if ( -d $tmpDir ) { executeOrDie("/bin/rm -rf ".$tmpDir); }
    executeOrDie("/bin/mkdir -pv ".$tmpDir."/SPECS ".$tmpDir."/SRPMS ".$tmpDir."/SOURCES");
    executeOrDie("/bin/cp -v ".$specFile." ".$tmpDir."/SPECS/");
    if ($specDir ne "spec") {
      ($main::verbosity > 1) && print "Creating working copy spec/\n";
      if (-e "spec") {
        cleanupAndDie("FATAL: There already is a file or directory called spec, something must have gone wrong.\n");
      }
      executeOrDie("/bin/cp -a ".$specDir." spec");
      if (! -d "spec") {
        cleanupAndDie("FATAL: /bin/cp -a ".$specDir." spec did not create a spec/ directory\n");
      }
    }
    my $tarCall = "/bin/tar cvzf ".$tmpDir."/SOURCES/".$specBase."-".$main::versionsBySpecFile{$specFile}.".tar.gz --xform 's%^./%".$specBase."-".$main::versionsBySpecFile{$specFile}."/%'";
    $tarCall .= " --exclude='debian*'";
    $tarCall .= " --exclude='spec-*'";
    $tarCall .= " --exclude='.svn'";
    $tarCall .= " --owner=root --group=root";
    $tarCall .= " .";
    if ($main::verbosity > 2) { print "tar call:\n".$tarCall."\n"; }
    if (!execute($tarCall)) {
      my $msg;
      if ($specDir ne "spec") {
        $msg = "Failed to make tarball for ".$specDir.", please investigate and then /bin/rm -rf spec/\n";
      } else {
        $msg = "Failed to make tarball ".$specDir.", please investigate spec/\n";
      }
      cleanupAndDie($msg);
    }
    if ($specDir ne "spec") {
      print "Cleaning up working copy spec/";
      executeOrDie("/bin/rm -rf spec");
    }
    my $rpmBuildCall = "/usr/bin/fakeroot /usr/bin/rpmbuild -bs --nodeps --define=\"_topdir ".$tmpDir."\" ".$tmpDir."/SPECS/".$specBase.".spec";
    # rpm signing is diffulcult, hence use signRpm() over --sign 
    if ($main::verbosity > 2) { print "rpmbuild call:\n".$rpmBuildCall."\n"; }
    executeOrDie($rpmBuildCall);
    # cleanupAndDie("Failed to rpmbuild ".$specDir." in ".$tmpDir", please investigate and then /bin/rm -rf ".$tmpDir"/\n");
    my $srpmName = $specBase."-".$main::versionsBySpecFile{$specFile}."-".$main::releasesBySpecFile{$specFile}.".src.rpm";
    my $mvCall = "/bin/mv -v ".$tmpDir."/SRPMS/".$srpmName." ..";
    executeOrDie($mvCall);

    # sign srpm
    if ($main::opts{"u"}) {
      print "Suppressing signature due to option -u\n";
    } else {
      print "Signing ../".$srpmName."\n";
      signRpm("../".$srpmName);
    }
 
    # push @generatedPackages
    print "Done building source package from ".$specFile."\n";

    if (-d $buildhost) {
      $feedScript .= proposeCpScript( $buildhost, "../".$srpmName );
      $releasedFilesDesc .= "From ".$specFile."/ to directory ".$buildhost."/:\n  "."  ../".$srpmName."\n\n";
    } else {
      my $uploadName = $main::specBase."_".$srpmName."_".$main::versionsBySpecFile{$specFile}."-".$main::releasesBySpecFile{$specFile};
      my $feedHostDir = $main::buildhostFeedHostDirs{$buildhost};
      if ($main::verbosity > 1) { print "buildHost=$buildhost, feedHostDir=$feedHostDir\n"; }
      $feedScript .= proposeFeedScript($feedHostDir, $uploadName, "../".$srpmName);
      $releasedFilesDesc .= "From ".$specFile."/ to buildhost ".$buildhost.":\n  "."  ../".$srpmName."\n\n";
    }
  } # end for specFile

  my $feedScriptFile = "/tmp/$$.sh";
  (-e $feedScriptFile) && unlink $feedScriptFile;
  $feedScript .= "\nexit 0\n";
  printToFile( $feedScriptFile, $feedScript, 0755 );
  if ($main::verbosity > 0) {
    print "Going to call feed script ".$feedScriptFile."\n";
  }
  executeOrDie($feedScriptFile);
  unlink $feedScriptFile;
  if ($main::verbosity > 0) {
    print "\nSuccessfully distributed these packages:\n\n".$releasedFilesDesc."\n";
  }

}


# checks if versions between used debian*/changelog and spec*/*.spec differ
# sets 
#   $main::topVersionVersion to top appVersion
#   $main::topVersionChangelog to debian*/changelog with top version
#   $main::versionDifferences to nr of differences
sub checkPackageVersions() {
  $main::topVersionVersion = "";
  $main::topVersionChangelog = "";
  $main::versionDifferences = 0;
  foreach my $changelogFile (@main::changelogFiles) {
    ($main::verbosity > 1) && print $changelogFile." is at version ".$main::appVersionByChangelog{$changelogFile}."\n";
    if ($main::topVersionVersion eq "") {
      $main::topVersionVersion = $main::appVersionByChangelog{$changelogFile};
      $main::topVersionChangelog = $changelogFile;
      next;
    }
    if ($main::topVersionVersion eq $main::appVersionByChangelog{$changelogFile}) {
      next;
    }
    # they differ, compare  
    $main::versionDifferences++;
    if (version_compare($main::appVersionByChangelog{$changelogFile}, $main::topVersionVersion) > 0) {
      $main::topVersionVersion = $main::appVersionByChangelog{$changelogFile};
      $main::topVersionChangelog = $changelogFile;
      print "WARNING: Version of ".$changelogFile.", ".$main::appVersionByChangelog{$changelogFile}." differs from top changelog version ".$main::topVersionVersion." from ".$main::topVersionChangelog."\n";
    }
  }
  foreach my $specFile (@main::specFiles) {
    ($main::verbosity > 1) && print $specFile." is at version ".$main::appVersionsBySpecFile{$specFile}."\n";
    if ($main::topVersionVersion eq $main::appVersionsBySpecFile{$specFile}) {
      next;
    }
    # they differ, compare  
    $main::versionDifferences++;
    if (version_compare($main::appVersionsBySpecFile{$specFile}, $main::topVersionVersion) > 0) {
      print "WARNING: Version of ".$specFile.", ".$main::appVersionsBySpecFile{$specFile}." differs from top changelog version ".$main::topVersionVersion." from ".$main::topVersionChangelog."\n";
    }
  }

}


# collects some infos in debian/devhost.txt.gz 

sub collectDevHostTxt() {
  if (-d "debian") {
    if (-f "debian/devhost.txt.gz") { 
      unlink "debian/devhost.txt.gz";
    }
    my $devHostInfo = "";
    $devHostInfo .= "Hostname: ".$ENV{"HOSTNAME"}."\n";
    $devHostInfo .= "User: ".$ENV{"USER"}."\n";
    $devHostInfo .= "Date: ".`LANG="C" date`."\n";
    if (-d ".svn") {
      $devHostInfo .= "svn info:\n".`LANG="C" svn info`;
    }
    $devHostInfo .= "dpkg -l:\n".`LANG="C" dpkg -l`."\n";
    printToFile("debian/devhost.txt", $devHostInfo);
    executeOrDie("gzip debian/devhost.txt");
    push @main::filesToCleanUp, "debian/devhost.txt.gz";
  }
}


# checks pom.xml content, for pba-checkpoms
# $1 = name of pom.xml file
sub checkPomXml( $ ) {
  my $pomXmlFile = shift;
  if (length($pomXmlFile) == 0) {
    cleanupAndDie("FATAL: checkPomXml called without pom.xml file.\n");
  }
  if (not -f $pomXmlFile) {
    cleanupAndDie("FATAL: checkPomXml called for non-existing file ".$pomXmlFile."\n");
  }
  ($main::verbosity > 2) && print "Checking pom ".$pomXmlFile."\n";
  my $svnUrl = qx(LANG=\"C\" svn info $pomXmlFile |egrep ^URL |cut -f2 -d' ');
  chomp $svnUrl;
  $svnUrl =~ s/\/pom.xml//;
  ($main::verbosity > 2) && print "SVN URL for directory is: ".$svnUrl."\n";
  if (length($svnUrl) == 0) {
    cleanupAndDie("FATAL: Unable to detect svn info for ".$pomXmlFile."\n");
  }
  my $correctScmConnection = "scm:svn:" . $svnUrl;
  my $correctScmDeveloperConnection = "scm:svn:" . $svnUrl;
  my $correctScmUrl = "";
  if ($main::useHttpForReadOnlyURLs) {
    $correctScmConnection =~ s/svn.ssh:\/\/[a-zA-Z0-9-]+@/http:\/\//;
    $correctScmUrl = $correctScmConnection;
    $correctScmUrl =~ s/scm:svn:http/http/;
  }
  if ($main::verbosity > 2) {
    print "Suggested scm connection: '".$correctScmConnection."'\n";
    print "Suggested scm developer connection: '".$correctScmDeveloperConnection."'\n";
    print "Suggested scm url: '".$correctScmUrl."'\n";
  }
  my $pomXmlContent = readCompleteFile($pomXmlFile);
  if ($pomXmlContent =~ /<scm>(.*)<\/scm>/s) {
    # investigate scm tag
    my $scmTag = $1;
    #print $pomXmlFile." has this scm tag:\n\n".$scmTag."\n\n";
    my $presentScmConnection = "";
    if ($scmTag =~ /<connection>(.*)<\/connection>/) {
      $presentScmConnection = $1; 
    }
    my $presentScmDeveloperConnection = "";
    if ($scmTag =~ /<developerConnection>(.*)<\/developerConnection>/) {
      $presentScmDeveloperConnection = $1; 
    }
    my $presentScmUrl = "";
    my $presentScmUrlTag = "";
    if ($scmTag =~ /<url>(.*)<\/url>/) {
      $presentScmUrl = $1; 
      $presentScmUrlTag = "<url>".$presentScmUrlTag."</url>";
    }
    if ($main::verbosity > 2) {
      print "Present scm connection: '".$presentScmConnection."'\n";
      print "Present scm developer connection: '".$presentScmDeveloperConnection."'\n";
      print "Present scm url: '".$presentScmUrl."'\n";
    }
    my $correctedPomXmlContent = $pomXmlContent;
    my $suggestionList = "";
    if ($correctScmConnection ne $presentScmConnection) {
      $suggestionList .= "Change scm connection from\n  ".$presentScmConnection."\nto\n  ".$correctScmConnection."\n\n";
      $presentScmConnection =~ s/\+/\\+/g; # escape + in svn+ssh, don't know why / needs no escaping
      $correctedPomXmlContent =~ s/<connection>$presentScmConnection<\/connection>/<connection>$correctScmConnection<\/connection>/;
    }
    if ($correctScmDeveloperConnection ne $presentScmDeveloperConnection) {
      $suggestionList .= "Change scm developer connection from\n  ".$presentScmDeveloperConnection."\nto\n  ".$correctScmDeveloperConnection."\n\n";
      $presentScmDeveloperConnection =~ s/\+/\\+/g; # escape + in svn+ssh, don't know why / needs no escaping
      $correctedPomXmlContent =~ s/<developerConnection>$presentScmDeveloperConnection<\/developerConnection>/<developerConnection>$correctScmDeveloperConnection<\/developerConnection>/;
    }
    if ($correctScmUrl ne $presentScmUrl) {
      if ($presentScmUrlTag eq "") {
        $suggestionList .= "Insert scm url tag\n  <url>".$correctScmUrl."</url>\n\n";
        $correctedPomXmlContent =~ s/<\/scm>/<url>$correctScmUrl<\/url><\/scm>/;
      } else {
        $suggestionList .= "Change scm url from\n  ".$presentScmUrl."\nto\n  ".$correctScmUrl."\n\n";
        $correctedPomXmlContent =~ s/<url>$presentScmUrl<\/url>/<url>$correctScmUrl<\/url>/;
      }
    }
    if ($suggestionList eq "") {
      print $pomXmlFile." looks GOOD.\n"
    } else {
      if ($main::applySugggestions gt 0) {
        print $pomXmlFile." looks BAD. Applying these changes:\n\n".$suggestionList."\n";
        printToFile( $pomXmlFile, $correctedPomXmlContent ); 
      } else {
        print $pomXmlFile." looks BAD, suggesting these changes:\n\n".$suggestionList."\n";
      }
    }
  } else {
    # no scm tag, suggest it
    print $pomXmlFile." looks BAD, it does not contain a scm tag. Suggesting to insert this:\n\n";
    print "  <scm>\n";
    print "    <connection>".$correctScmConnection."</connection>\n";
    print "    <developerConnection>".$correctScmDeveloperConnection."</developerConnection>\n";
    if ($main::useHttpForReadOnlyURLs) {
      print "    <url>".$correctScmUrl."</url>\n";
    }
    print "  </scm>\n";
    print "\n\n";
  }

}



# EOF
1;

